#include <string>
#include <vector>
#include <iostream>
#include <algorithm>
#include <cstdio>
#include <set>

static bool maysort = true;

/* Characters can never be split. */
typedef std::string Character;
typedef std::vector<Character> Item;
typedef std::vector<Item> Itemlist;

static const Item Import(const std::string& s)
{
    Item result(s.size());
    for(unsigned a=0; a<s.size(); ++a)
    {
        Character ch;
        switch(s[a])
        {
            case ' ':
            case '\'': case '"':
            case ',': case '$':
            case '{': case '}':
            case '&': case '|':
            case '(': case ')':
            case '\\': ch += '\\'; //passthru
            default: ch += s[a];
        }
        result[a] = ch;
    }
    return result;
}
static const std::string Export(const Item& item)
{
    std::string result;
    for(unsigned a=0; a<item.size(); ++a)
        result += item[a];
    return result;
}

static const Item Characterify(const Itemlist& items)
{
    std::string token = "{";
    for(unsigned a=0; a<items.size(); ++a)
    {
        if(a > 0) token += ',';
        token += Export(items[a]);
    }
    token += "}";

    Item result;
    result.push_back(token);
    return result;
}

static const Item Slice(const Item& item, unsigned pos, unsigned length)
{
    return Item(item.begin() + pos,
                item.begin() + pos+length);
}
static const Item SliceBegin(const Item& item, unsigned length)
{
    if(length >= item.size()) return item;
    return Slice(item, 0, length);
}
static const Item SliceEnd(const Item& item, unsigned length)
{
    if(length >= item.size()) return item;
    return Slice(item, item.size()-length, length);
}

static const Itemlist Pack(Itemlist lines)
{

Restart:
    unsigned max_length = 0;
    for(unsigned a=0; a<lines.size(); ++a)
        max_length = std::max(max_length, (unsigned)lines[a].size());
    
    unsigned best_winning = 0;
    unsigned best_begin   = 0;
    unsigned best_end     = 0;
    
    for(unsigned begin=0; begin<max_length; ++begin)
        for(unsigned end=!begin; end+begin < max_length; ++end)
        {
            std::set<unsigned> ignore;
            
            for(unsigned a=0; a<lines.size(); ++a)
            {
                if(ignore.find(a) != ignore.end()) continue;
                if(lines[a].size() < begin+end) continue;
                
                Item beginpart = SliceBegin(lines[a], begin);
                Item endpart   = SliceEnd(lines[a],   end);
                
                unsigned winning_per_one
                    = (Export(beginpart).size() + Export(endpart).size()) * max_length
                    + std::max(begin, end);
                
                if(beginpart.size() != begin
                || endpart.size() != end)
                {
                    fprintf(stderr, "eeh? %u is not %u or %u is not %u (total %u)\n",
                        (unsigned)beginpart.size(), begin,
                        (unsigned)endpart.size(), end,
                        (unsigned)lines[a].size());
                }
                
                std::vector<unsigned> matchlist;
                
                __label__ nomatch;
                
                for(unsigned b=a+1; b<lines.size(); ++b)
                {
                    if(lines[b].size() >= begin+end)
                    {
                        unsigned endoffs = lines[b].size() - end;
                        for(unsigned c=0; c<lines[b].size(); ++c)
                        {
                            if(c < begin) { if(lines[b][c] != beginpart[c]) goto nomatch; continue; }
                            if(c == begin) { if(!end) break; c = endoffs; }
                            
                            if(lines[b][c] != endpart[c-endoffs]) goto nomatch;
                        }
                        // match.
                        matchlist.push_back(b);
                        continue;
                    }
                  nomatch: ;
                    if(!maysort) break;
                }
                
                unsigned num_matches = matchlist.size();
                if(num_matches == 0) continue;
                
                unsigned winning = winning_per_one * (num_matches+1);
                if(winning > best_winning)
                {
                    best_winning = winning;
                    best_begin   = begin;
                    best_end     = end;
                }
                ignore.insert(matchlist.begin(), matchlist.end());
            }
        }

    if(best_winning > 0)
    {
        unsigned begin = best_begin;
        unsigned end   = best_end;

        std::set<unsigned> ignore;
        
        for(unsigned a=0; a<lines.size(); ++a)
        {
            if(ignore.find(a) != ignore.end()) continue;
            if(lines[a].size() < begin+end) continue;
            
            Item beginpart = SliceBegin(lines[a], begin);
            Item endpart   = SliceEnd(lines[a], end);
            
            unsigned winning_per_one
                = (Export(beginpart).size() + Export(endpart).size()) * max_length
                + std::max(begin, end);
            
            std::vector<unsigned> matchlist;
            matchlist.push_back(a);
            
            __label__ nomatch;

            for(unsigned b=a+1; b<lines.size(); ++b)
            {
                if(lines[b].size() >= begin+end)
                {
                    unsigned endoffs = lines[b].size() - end;
                    for(unsigned c=0; c<lines[b].size(); ++c)
                    {
                        if(c < begin) { if(lines[b][c] != beginpart[c]) goto nomatch; continue; }
                        if(c == begin) { if(!end) break; c = endoffs; }
                        if(lines[b][c] != endpart[c-endoffs]) goto nomatch;
                    }
                    // match.
                    matchlist.push_back(b);
                    continue;
                }
              nomatch: ;
                if(!maysort) break;
            }
            
            unsigned num_matches = matchlist.size() - 1;
            // -1 because a was included this time
            if(num_matches == 0) continue;
            
            unsigned winning = winning_per_one * (num_matches+1);
            if(winning == best_winning)
            {
                /*
                printf("Yay, beginpart='%s'(%u) endpart='%s'(%u), num matches=%u\n",
                    Export(beginpart).c_str(),begin,
                    Export(endpart).c_str(),end,
                    (unsigned)matchlist.size());
                */
                Itemlist slices;
                
                for(unsigned b=0; b<matchlist.size(); ++b)
                {
                    const Item& line = lines[matchlist[b]];
                    slices.push_back(Slice(line, begin, line.size()-end-begin));
                }
                for(unsigned b=matchlist.size()-1; b>0; --b)
                    lines.erase(lines.begin() + matchlist[b]);
                // the item at "a" is still there.
                
                /*printf("      middlepart_in='%s'\n",
                    Export(Characterify(slices)).c_str());
                */
                
                slices = Pack(slices);
                Item middlepart = Characterify(slices);

                /*
                printf("      middlepart_out='%s'\n",
                    Export(middlepart).c_str());
                */
                
                lines[a].clear();
                lines[a].insert(lines[a].end(), beginpart.begin(), beginpart.end());
                lines[a].insert(lines[a].end(), middlepart.begin(),middlepart.end());
                lines[a].insert(lines[a].end(), endpart.begin(),   endpart.end());
                goto Restart;
            }
            ignore.insert(matchlist.begin(), matchlist.end());
        }
    }
    
    return lines;
}


//#define _GNU_SOURCE
#include <unistd.h>
#include <getopt.h>

int main(int argc,char**argv)
{
    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"help",     0, 0,'h'},
            {"version",  0, 0,'V'},
            {"nosort",   0, 0,'n'},
            {0,0,0,0}
        };
        int c = getopt_long(argc, argv, "hVn", long_options, &option_index);
        if(c==-1) break;
        switch(c)
        {
            case 'V':
            {
                std::cout << VERSION"\n";
                return 0;
            }
            case 'h':
            {
                std::cout <<
                    "shellpack v"VERSION" - Copyright (C) 1992,2006 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\n"
                    "Usage: shellpack [<options>] < wordlist\n"
                    "\n"
                    "Transforms a wordlist into a compressed expression that can be used\n"
                    "in a sh-compatible shell command line.\n"
                    "\n"
                    " --help, -h     This help\n"
                    " --version, -V  Displays version information\n"
                    " --nosort, -n   Retains the original order of words (faster, but less powerful)\n"
                    "\n";
                return 0;
            }
            case 'n':
                maysort = false;
                break;
            default: goto argerror;
        }
    }
    
    if(argc != optind)
    {
    argerror:
        std::cerr << "shellpack: Invalid parameters. Try `shellpack --help'\n";
        return 1;
    }

    Itemlist lines;

    for(;;) 
    {
        std::string s;
        std::getline(std::cin, s);
        if(!std::cin.good())break;
        lines.push_back(Import(s));
    }
    if(maysort)
        std::sort(lines.begin(), lines.end());

    lines = Pack(lines);
    
    for(unsigned a=0; a<lines.size(); ++a)
    {
        if(a>0) std::printf(" ");
        std::printf("%s", Export(lines[a]).c_str());
    }
    std::printf("\n");
    
    return 0;
}
